# @Time     : 2021/8/12
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from datetime import datetime
from enum import IntEnum

from ..time_ import UnitOfTime, TimeType
from ..time_.api import timestamp


class StartStates(IntEnum):
    """
    运行状态
    """
    UN_START = 0  # 停
    STARTED = 1  # 已开始
    PAUSE = 2  # 暂停


class IMeterTimer(metaclass=ABCMeta):
    """
    消息计时器
    """

    @property
    @abstractmethod
    def is_timeout(self) -> bool:
        """
        is_timeout
        :return:
        """
        pass

    @abstractmethod
    def current(self) -> float:
        """
        get current time
        :return:
        """
        pass

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def restart(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        """
        stop timer
        :return:
        """
        pass

    @abstractmethod
    def clear(self) -> None:
        """
        clear timeout
        :return:
        """
        pass


class MeterTimer(IMeterTimer):

    def __init__(self, timeout: TimeType, unt: UnitOfTime):
        self._st: TimeType = 0  # 开始时间
        self._state: StartStates = StartStates.UN_START  # 状态

        self._timestamp = timestamp(unt)  # 时间单位
        self._timeout = timeout  # 超时时间

    @property
    def is_started(self) -> bool:
        return self._state == StartStates.STARTED

    @property
    def is_timeout(self) -> bool:
        return self._is_timeout()

    def current(self) -> float:
        return self._now_timestamp()

    def start(self) -> None:
        if self.is_started:
            return
        # update state, start time
        self._update_state(StartStates.STARTED)
        self._update_start(self.current())

    def restart(self) -> None:
        self._update_state(StartStates.STARTED)
        self._update_start(self.current())

    def stop(self) -> None:
        self._update_state(StartStates.PAUSE)

    def clear(self) -> None:
        self._update_state(StartStates.UN_START)
        self._update_start(0)

    def _update_start(self, st: TimeType):
        self._st = st

    def _update_state(self, state: StartStates):
        self._state = state

    def _now_timestamp(self):
        return datetime.now().timestamp()

    def _is_timeout(self):
        return self._timestamp.convert(self.current() - self._st) >= self._timeout

    def __str__(self):
        return f"{self.__class__.__name__}(timeout={self._timeout}, state={self._state})"



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import UnitOfTime
from .unit import Unt, MulOrTruediv
from .timestamp import Timestamp
from .timeunt import TimeUnt

DEFAULT = UnitOfTime.MILLISECOND


def second(mul_or_truediv: MulOrTruediv = MulOrTruediv.MUL):
    return Unt(UnitOfTime.SECOND, mul_or_truediv)


def microsecond(mul_or_truediv: MulOrTruediv = MulOrTruediv.MUL):
    return Unt(UnitOfTime.MICROSECOND, mul_or_truediv)


def millisecond(mul_or_truediv: MulOrTruediv = MulOrTruediv.MUL):
    return Unt(UnitOfTime.MILLISECOND, mul_or_truediv)


def timestamp(unit_of_time: UnitOfTime = DEFAULT):
    return Timestamp(unit_of_time)


def timeunt(unit_of_time: UnitOfTime = DEFAULT):
    return TimeUnt(unit_of_time)



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from enum import IntEnum
from typing import Union

IntOrFloat = Union[int, float]
_B = 10


class UnitOfTime(IntEnum):
    SECOND = 1  # 秒
    MILLISECOND = _B ** 3  # 毫秒
    MICROSECOND = _B ** 6  # 微秒


class AbstractTimeConvert(metaclass=ABCMeta):

    def __init__(self, unit_of_time: UnitOfTime):
        self._unit_of_time = unit_of_time

    @property
    def unit_of_time(self):
        return self._unit_of_time

    @unit_of_time.setter
    def unit_of_time(self, value: UnitOfTime):
        if isinstance(value, UnitOfTime):
            self._unit_of_time = value

    def convert(self, value: IntOrFloat) -> float:
        return self._convert(value)

    @abstractmethod
    def _convert(self, value: IntOrFloat) -> float:
        pass

    def __str__(self):
        return f"{self.__class__.__name__}(unit_of_time={self.unit_of_time})"



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from ..utils import PathTypes, EncodingType
from .factory import ReaderFactory


def conf_reader(file: PathTypes, encoding: EncodingType = None):
    return ReaderFactory.create(file, encoding)



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .conf.api import conf_reader
from .time_.api import timeunt, timestamp, second, millisecond, microsecond
from .meter.api import meter
from .validator.api import check_function, check_async_function
from .callback.api import func_manager, afunc_manager
from .task.api import executor, creat_task



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from asyncio import Future, AbstractEventLoop, get_event_loop
from operator import not_
from typing import Callable


class AbstractFutConvert(metaclass=ABCMeta):

    def __init__(self, eloop: AbstractEventLoop = None):
        self._eloop = eloop

    @property
    def eloop(self):
        return self._eloop

    @property
    @abstractmethod
    def executor(self):
        pass

    def convert(self, sync: Callable, *args) -> Future:
        return self._to_fut(sync, *args)

    def _to_fut(self, sync: Callable, *args):
        eloop = self._get_event_loop()
        return eloop.run_in_executor(self.executor, sync, *args)

    def _get_event_loop(self):
        if not_(self._eloop):
            self._eloop = get_event_loop()
        return self._eloop

    def __str__(self):
        return f"{self.__class__.__name__}(loop={self._eloop}, exec={self.executor})"


class FutConvert(AbstractFutConvert):

    def __init__(self, eloop: AbstractEventLoop = None, executor=None):
        super().__init__(eloop)
        self._executor = executor

    @property
    def executor(self):
        return self._executor



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import abstractmethod
from operator import not_, indexOf
from typing import Callable, List
from .base import AbstractCallbackManager
from ..validator.api import check_function, check_async_function

ListFunction = List[Callable]


class ListCallbackManager(AbstractCallbackManager):

    def __init__(self):
        self._callbacks: ListFunction = []

    @property
    def callbacks(self):
        return self._callbacks

    @property
    def length(self) -> int:
        return len(self._callbacks)

    def add(self, function: Callable) -> int:
        # self.check(function)
        if not self._is_exits(function):
            self._callbacks.append(function)
        return self.length

    def remove(self, function: Callable) -> int:
        if self.empty():
            return self.length
        idx = self._find(function)
        if idx > -1:
            self._callbacks.pop(idx)
        return self.length

    def _is_exits(self, function: Callable) -> bool:
        if not_(self._callbacks):
            return False
        return function in self._callbacks

    def _find(self, function) -> int:
        try:
            return indexOf(self._callbacks, function)
        except ValueError:
            return -1


class CallbackManagerDecorate(AbstractCallbackManager):

    def __init__(self, manager: AbstractCallbackManager):
        self._manager = manager

    @property
    def manager(self):
        return self._manager

    @manager.setter
    def manager(self, manager: AbstractCallbackManager):
        if isinstance(manager, AbstractCallbackManager):
            self._manager = manager

    @property
    def callbacks(self):
        return self.manager.callbacks

    @property
    def length(self) -> int:
        return self.manager.length

    def add(self, function: Callable) -> int:
        self._check(function)
        return self.manager.add(function)

    def remove(self, function: Callable) -> int:
        return self.manager.remove(function)

    @abstractmethod
    def _check(self, function: Callable) -> bool:
        pass


class SyncCallbackManager(CallbackManagerDecorate):

    def __init__(self, manager: AbstractCallbackManager):
        super().__init__(manager)

    def _check(self, value):
        return check_function(value)


class AsyncCallbackManager(CallbackManagerDecorate):

    def __init__(self, manager: AbstractCallbackManager):
        super().__init__(manager)

    def _check(self, value):
        return check_async_function(value)



# @Time     : 2021/8/12
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from typing import Callable

from ..is_ import is_validate
from ..time_ import UnitOfTime, TimeType
from .counter import MeterCounter
from .timer import MeterTimer

Function = Callable
INITIAL = 0


class AbstractMeter(metaclass=ABCMeta):

    def __init__(self, timeout: TimeType, unt: UnitOfTime = UnitOfTime.MILLISECOND):
        self._counter = MeterCounter(INITIAL)
        self._timer = MeterTimer(timeout, unt)

    def is_timeout(self):
        return self._timer.is_timeout

    def start(self):
        self._timer.start()

    def restart(self):
        self._timer.restart()
        self._counter.clear()

    def auto_increment(self):
        self._counter.auto_increment()

    def trigger(self, cb: Function = None) -> None:
        if is_validate.is_function(cb):
            cb(self._counter)

    @abstractmethod
    def tracking(self, func: Function) -> Function:
        pass

    def __str__(self):
        return f"{self.__class__.__name__}(timer={self._timer}, counter={self._counter})"

    def __call__(self, fn: Function):
        return self.tracking(fn)



# @Time     : 2021/8/12
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import iscoroutine, iscoroutinefunction, isfuture, Task
from collections.abc import Iterable, Callable
from enum import Enum
from operator import is_, is_not, not_
from pathlib import Path


class is_validate(object):
    PY_OBJ = (
        str,
        int, float,
        bool,
        dict,
        list, tuple
    )

    @staticmethod
    def is_string(value):
        return isinstance(value, str)

    @staticmethod
    def is_bstring(value):
        return isinstance(value, bytes)

    @staticmethod
    def is_bstring_array(value):
        return isinstance(value, bytearray)

    @staticmethod
    def is_any_string(value):
        return any([
            is_validate.is_string(value),
            is_validate.is_bstring(value)
        ])

    @staticmethod
    def is_boolean(value):
        return isinstance(value, bool)

    @staticmethod
    def is_list(value):
        return isinstance(value, list)

    @staticmethod
    def is_set(value):
        return isinstance(value, set)

    @staticmethod
    def is_tup(value):
        return isinstance(value, tuple)

    @staticmethod
    def is_dict(value):
        return isinstance(value, dict)

    @staticmethod
    def is_enum(value):
        return isinstance(value, Enum)

    @staticmethod
    def is_null(value):
        return is_(value, None)

    @staticmethod
    def is_not_null(value):
        return is_not(value, None)

    @staticmethod
    def is_int(value):
        return isinstance(value, int)

    @staticmethod
    def is_float(value):
        return isinstance(value, float)

    @staticmethod
    def is_number(value):
        return any([
            is_validate.is_int(value),
            is_validate.is_float(value)
        ])

    @staticmethod
    def is_path(value):
        return isinstance(value, Path)

    @staticmethod
    def is_iterable(value):
        return isinstance(value, Iterable)

    @staticmethod
    def is_function(value):
        return isinstance(value, Callable)

    @staticmethod
    def is_coroutine(value):
        return iscoroutine(value)

    @staticmethod
    def is_async_function(value):
        return iscoroutinefunction(value)

    @staticmethod
    def is_fut(value):
        return isfuture(value)

    @staticmethod
    def is_coroutine_task(value):
        return isinstance(value, Task)

    @staticmethod
    def is_py_empty(value):
        return not_(value)

    @staticmethod
    def is_py_obj(value):
        return isinstance(value, is_validate.PY_OBJ)



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import ITaskExecutor, ITaskFactory, ArgsTypes, KwargsTypes, CoroOrFunction
from .executor import TaskExecutor
from .factory import TaskFactory



# @Time     : 2021/6/2
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import AbstractEventLoop
from .base import AbstractTaskFactory, ArgsTypes, CoroOrFunction, KwargsTypes, _get_event_loop
from .adapters import CoroutineAdapters
from ..validator.is_validators import is_coroutine, is_async_function, is_function


class CoroTaskFactory(AbstractTaskFactory):

    def create_task(self):
        return self._create_by_coro(self.coro_or_func)


class CoroFunctionTaskFactory(AbstractTaskFactory):

    def create_task(self):
        """
        async function to task

        :return:
        """
        return self._create_by_coro(self._run_in_async_ctx())

    async def _run_in_async_ctx(self):
        """
        async function must be async context
        :return:
        """
        args, kwargs = self._args, self._kwargs
        return await self.coro_or_func(*args, **kwargs)


class FunctionTaskFactory(AbstractTaskFactory):

    def __init__(self,
                 loop: AbstractEventLoop,
                 *,
                 coro_or_func: CoroOrFunction,
                 args: ArgsTypes = None,
                 kwargs: KwargsTypes = None,
                 max_workers: int = None):
        super().__init__(coro_or_func, args, kwargs)
        self._eloop = loop
        self._max_workers = max_workers

    @property
    def eloop(self):
        return self._eloop

    def create_task(self):
        """
        function to task
        :return:
        """
        return self._create_by_coro(self._run_in_async_ctx())

    async def _run_in_async_ctx(self):
        fut = self._to_future(
            self.eloop,
            func=self.coro_or_func,
            args=self._args,
            kwargs=self._kwargs,
            max_workers=self._max_workers
        )
        return await fut

    def _to_future(self,
                   loop: AbstractEventLoop,
                   *,
                   func: CoroOrFunction,
                   args: tuple,
                   kwargs: dict,
                   max_workers: int = None):
        adapter = CoroutineAdapters(loop, func, max_workers=max_workers)
        return adapter.to_coroutine(*args, **kwargs)


class TaskFactory(AbstractTaskFactory):

    def __init__(self,
                 coro_or_func: CoroOrFunction,
                 args: ArgsTypes = None,
                 kwargs: KwargsTypes = None,
                 max_workers: int = None):
        super().__init__(coro_or_func, args, kwargs)
        self._max_workers = max_workers

    def create_task(self):
        coro_or_func = self.coro_or_func
        if is_coroutine(coro_or_func):
            return self._use_coro_factory(coro_or_func)
        elif is_async_function(coro_or_func):
            return self._use_async_factory(coro_or_func)
        elif is_function(coro_or_func):
            return self._use_func_factory(coro_or_func)
        else:
            raise ValueError(
                f"coro_or_func need a Coroutine, async function or function, but got {type(coro_or_func).__name__}"
            )

    def _use_coro_factory(self, coro_or_func: CoroOrFunction):
        args, kwargs = self._args, self._kwargs
        coro = CoroTaskFactory(coro_or_func, args, kwargs)
        return coro.create_task()

    def _use_async_factory(self, coro_or_func: CoroOrFunction):
        args, kwargs = self._args, self._kwargs
        async_func = CoroFunctionTaskFactory(coro_or_func, args, kwargs)
        return async_func.create_task()

    def _use_func_factory(self, coro_or_func: CoroOrFunction):
        event_loop = _get_event_loop()
        func = FunctionTaskFactory(
            event_loop,
            coro_or_func=coro_or_func,
            args=self._args, kwargs=self._kwargs,
            max_workers=self._max_workers
        )
        return func.create_task()



# @Time     : 2021/8/12
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from functools import wraps

from ..time_ import UnitOfTime, TimeType
from .base import Function, AbstractMeter


class AsyncMessageMeter(AbstractMeter):

    def __init__(self,
                 timeout: TimeType,
                 unit: UnitOfTime = UnitOfTime.MILLISECOND,
                 *,
                 callback: Function = None):
        super().__init__(timeout, unit)
        self._callback = callback

    @property
    def callback(self):
        return self._callback

    def tracking(self, func: Function) -> Function:
        meter = self

        @wraps(func)
        async def async_decorate(*args, **kwargs):
            meter.start()
            meter.auto_increment()
            result = await func(*args, **kwargs)
            if meter.is_timeout():
                meter.trigger(meter.callback)
                meter.restart()
            return result

        return async_decorate



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum
from operator import mul, truediv

from .base import AbstractTimeConvert, UnitOfTime, IntOrFloat


class MulOrTruediv(Enum):
    MUL = "mul"
    TRUEDIV = "truediv"


class Unt(AbstractTimeConvert):
    _map = {MulOrTruediv.MUL: mul, MulOrTruediv.TRUEDIV: truediv}

    def __init__(self, unit_of_time: UnitOfTime, mul_or_truediv: MulOrTruediv):
        super().__init__(unit_of_time)
        self._mul_or_truediv = mul_or_truediv

    def _convert(self, value: IntOrFloat) -> float:
        _math = self._get_mul_or_truediv()
        # print(_math)
        return _math(value, self.unit_of_time.value)

    def _get_mul_or_truediv(self):
        return self._map.get(self._mul_or_truediv, MulOrTruediv.MUL)


def unt_map(mul_or_truediv: MulOrTruediv = MulOrTruediv.MUL):
    return {
        UnitOfTime.SECOND: Unt(UnitOfTime.SECOND, mul_or_truediv),
        UnitOfTime.MILLISECOND: Unt(UnitOfTime.MILLISECOND, mul_or_truediv),
        UnitOfTime.MICROSECOND: Unt(UnitOfTime.MICROSECOND, mul_or_truediv)
    }



# @Time     : 2021/8/12
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod

MIN = 0
EACH = 1


class IMeterCounter(metaclass=ABCMeta):
    """
    消息计数器
    """

    @abstractmethod
    def current(self) -> int:
        pass

    @abstractmethod
    def auto_increment(self) -> None:
        """
        auto_increment
        :return:
        """
        pass

    @abstractmethod
    def clear(self) -> None:
        """
        clear
        :return:
        """
        pass


class MeterTotal(IMeterCounter):

    def __init__(self, initial: int = 0):
        if initial <= 0:
            self._total = MIN
        else:
            self._total = initial

    def current(self) -> int:
        return self._total

    def auto_increment(self) -> None:
        self._increment(EACH)

    def clear(self) -> None:
        """
        clear count
        :return:
        """
        self._total = 0

    def _increment(self, value: int) -> None:
        self._total += value

    def __str__(self):
        return f"{self.__class__.__name__}(current={self.current()})"


class MeterCurrent(IMeterCounter):

    def __init__(self):
        self._current = 0

    def current(self) -> int:
        return self._current

    def auto_increment(self) -> None:
        self._increment(EACH)

    def clear(self) -> None:
        """
        clear count
        :return:
        """
        self._current = 0

    def _increment(self, value: int) -> None:
        self._current += value

    def __str__(self):
        return f"{self.__class__.__name__}(current={self.current()})"


class MeterCounter(IMeterCounter):

    def __init__(self, initial: int = 0):
        self._total = MeterTotal(initial)  # 累计
        self._current = MeterCurrent()  # 当前

    def current(self) -> int:
        return self._current.current()

    def total(self) -> int:
        return self._total.current()

    def auto_increment(self) -> None:
        self._total.auto_increment()
        self._current.auto_increment()

    def clear(self) -> None:
        self._current.clear()

    def __str__(self):
        return f"{self.__class__.__name__}(current={self._current}, total={self._total})"



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import AbstractEventLoop, create_task
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from operator import not_

from ..is_ import is_validate
from .base import ITaskFactory, CoroOrFunction, ArgsTypes, KwargsTypes
from .fut import FutConvert


def _create_task(coro):
    if not is_validate.is_coroutine(coro):
        raise ValueError(
            f"coro need Coroutine, but got {type(coro).__name__}"
        )
    return create_task(coro)


class CoroTaskFactory(ITaskFactory):

    def __init__(self, coro_or_func: CoroOrFunction):
        self._coro = coro_or_func

    def create(self):
        return _create_task(self._coro)

    def __str__(self):
        return f"{self.__class__.__name__}(coro={self._coro})"


class CoroFunctionTaskFactory(ITaskFactory):

    def __init__(self, coro_or_func: CoroOrFunction, *, args: ArgsTypes, kwargs: KwargsTypes):
        self._afunc = coro_or_func
        self._args, self._kwargs = args, kwargs

    def create(self):
        """
        async function to task

        :return:
        """
        return _create_task(self._run_in_ctx())

    async def _run_in_ctx(self):
        """
        async function must be async context
        :return:
        """
        args, kwargs = self._args, self._kwargs
        return await self._afunc(*args, **kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}(afunc={self._afunc})"


class FunctionTaskFactory(ITaskFactory):

    def __init__(self,
                 coro_or_func: CoroOrFunction,
                 *,
                 eloop: AbstractEventLoop,
                 max_workers: int = None,
                 args: ArgsTypes = None,
                 kwargs: KwargsTypes = None):
        self._func = coro_or_func
        self._args, self._kwargs = args, kwargs
        self._convert = FutConvert(
            eloop, ThreadPoolExecutor(max_workers=max_workers)
        )

    def create(self):
        """
        function to task
        :return:
        """
        return _create_task(self._run_in_ctx())

    async def _run_in_ctx(self):
        sync = self._get_sync_func()
        fut = self._to_future(sync)
        return await fut

    def _to_future(self, sync):
        return self._convert.convert(sync, *self._args)

    def _get_sync_func(self):
        return partial(self._func, **self._kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}(func={self._func})"


class TaskFactory(ITaskFactory):

    def __init__(self,
                 coro_or_func: CoroOrFunction,
                 *,
                 eloop: AbstractEventLoop,
                 max_workers: int = None,
                 args: ArgsTypes = None,
                 kwargs: KwargsTypes = None):
        _args = () if not_(args) else args
        _kwargs = {} if not_(kwargs) else kwargs

        if is_validate.is_coroutine(coro_or_func):
            self._factory = CoroTaskFactory(coro_or_func)

        elif is_validate.is_async_function(coro_or_func):
            self._factory = CoroFunctionTaskFactory(coro_or_func, args=_args, kwargs=_kwargs)

        elif is_validate.is_function(coro_or_func):
            self._factory = FunctionTaskFactory(
                coro_or_func,
                eloop=eloop, max_workers=max_workers,
                args=_args, kwargs=_kwargs
            )
        else:
            raise ValueError(
                f"coro_or_func need a Coroutine, async function or function, but got {type(coro_or_func).__name__}"
            )

    @property
    def factory(self):
        return self._factory

    def create(self):
        return self.factory.create()

    def __str__(self):
        return f"{self.__class__.__name__}(factory={self.factory})"



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from operator import not_
from typing import Callable, Iterable


class AbstractCallbackManager(metaclass=ABCMeta):

    @property
    @abstractmethod
    def callbacks(self) -> Iterable[Callable]:
        pass

    @property
    @abstractmethod
    def length(self) -> int:
        pass

    def empty(self):
        return not_(self.length)

    @abstractmethod
    def add(self, function: Callable) -> int:
        pass

    @abstractmethod
    def remove(self, function: Callable) -> int:
        pass

    def to_list(self):
        if self.empty():
            return []
        return [cb for cb in self]

    def __iter__(self):
        if not self.empty():
            for _, cb in enumerate(self.callbacks):
                yield cb

    # def is_callback_manager(self, value):
    #     return isinstance(value, AbstractCallbackManager)

    def __str__(self):
        return f"{self.__class__.__name__}(callbacks={[item for item in self]})"



# @Time     : 2021/8/12
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from typing import Union
from enum import IntEnum

IntOrFloat = Union[int, float]
_BASE = 10


class UnitOfTime(IntEnum):
    SECOND = _BASE ** 0  # 秒
    MILLISECOND = _BASE ** 3  # 毫秒
    MICROSECOND = _BASE ** 6  # 微秒


class ITimestampConvert(metaclass=ABCMeta):

    @abstractmethod
    def convert(self, timestamp: IntOrFloat) -> float:
        pass


class SecondTimestamp(ITimestampConvert):

    def convert(self, timestamp: IntOrFloat) -> float:
        return timestamp * UnitOfTime.SECOND.value


class MilliSecondTimestamp(ITimestampConvert):

    def convert(self, timestamp: IntOrFloat) -> float:
        return timestamp * UnitOfTime.MILLISECOND.value


class MicroSecondTimestamp(ITimestampConvert):

    def convert(self, timestamp: IntOrFloat) -> float:
        return timestamp * UnitOfTime.MICROSECOND.value


class Timestamp(ITimestampConvert):
    _map = {
        UnitOfTime.SECOND: SecondTimestamp(),
        UnitOfTime.MILLISECOND: MilliSecondTimestamp(),
        UnitOfTime.MICROSECOND: MicroSecondTimestamp()
    }

    def __init__(self, unit_of_time: UnitOfTime):
        self._unit_of_time = unit_of_time

    @property
    def unit_of_time(self):
        return self._unit_of_time

    @unit_of_time.setter
    def unit_of_time(self, value):
        if isinstance(value, UnitOfTime):
            self._unit_of_time = value

    def convert(self, timestamp: IntOrFloat) -> float:
        return self._get_convert().convert(float(timestamp))

    def __str__(self):
        return f"{self.__class__.__name__}(unit_of_time={self.unit_of_time})"

    def _get_convert(self):
        return self._map.get(self.unit_of_time, UnitOfTime.MILLISECOND)


def timestamp(unit_of_time: UnitOfTime = UnitOfTime.MILLISECOND):
    return Timestamp(unit_of_time)



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import AbstractEventLoop

from .base import CoroOrFunction, ArgsTypes, KwargsTypes
from .executor import TaskExecutor
from .factory import TaskFactory


def executor(coro_or_func: CoroOrFunction, *, eloop: AbstractEventLoop = None, max_workers: int = None):
    return TaskExecutor(coro_or_func, eloop=eloop, max_workers=max_workers)


def creat_task(coro_or_func: CoroOrFunction,
               *,
               eloop: AbstractEventLoop = None,
               max_workers: int = None,
               args: ArgsTypes = None,
               kwargs: KwargsTypes = None):
    return TaskFactory(
        coro_or_func, eloop=eloop, max_workers=max_workers,
        args=args, kwargs=kwargs
    ).create()



# @Time     : 2021/5/31
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from ..utils import StringsUtil
from ..is_ import is_validate


class ReaderUtil(object):

    @staticmethod
    def to_int(value):
        if is_validate.is_int(value):
            return value
        elif is_validate.is_number(value):
            return int(value)
        else:
            raise ValueError(f"value need float or int, but got {type(value).__name__}")

    @staticmethod
    def to_float(value):
        if is_validate.is_float(value):
            return value
        elif is_validate.is_number(value):
            return float(value)
        else:
            raise ValueError(f"value need float or int, but got {type(value).__name__}")

    @staticmethod
    def to_string(value):
        if is_validate.is_string(value):
            return value
        elif is_validate.is_any_string(value):
            return StringsUtil.anystr_to_string(value)
        return str(value)



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractTimeConvert, UnitOfTime, IntOrFloat as TimeType
from .unit import UnitOfTime, MulOrTruediv, Unt, unt_map
from .timestamp import Timestamp
from .timeunt import TimeUnt



# @Time     : 2021/6/2
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import (
    ICoroutineAdapters,
    AbstractExecutor,
    AbstractTaskFactory,
    IAsyncTask,
    ArgsTypes,
    CoroOrFunction,
    KwargsTypes,
    ReturnType,
    _get_event_loop
)
from .adapters import CoroutineAdapters
from .executor import (
    CoroExecutor,
    CoroFunctionExecutor,
    FunctionExecutor,
    Executor
)
from .task import (
    CoroTaskFactory,
    CoroFunctionTaskFactory,
    FunctionTaskFactory,
    TaskFactory
)



# @Time     : 2021/6/1
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from ..is_ import is_validate
from ..error import NotFunctionError, NotAsyncFunctionError
from .base import AbstractValidator


class FunctionValidator(AbstractValidator):

    def is_validate(self, value) -> bool:
        return is_validate.is_function(value)

    def raise_error(self, message: str):
        raise NotFunctionError(message)


class AsyncFunctionValidator(AbstractValidator):

    def is_validate(self, value) -> bool:
        return is_validate.is_async_function(value)

    def raise_error(self, message: str):
        raise NotAsyncFunctionError(message)



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
# from .src.is_ import is_validate
# from .src.validator.error.base import BaseError
# from .src.validator.error.errors import NotFunctionError, NotAsyncFunctionError
# from .src.validator.validators.base import AbstractValidator
# from .src.validator.validators.func import FunctionValidator
# from .src.validator.validators.afunc import AsyncFunctionValidator
# from .src.validator.manager.base import IValidatorManager, ValidatorTypes
# from .src.validator.manager.managers import ValidatorManager, ValidatorList
# from .src.validator.api import (
#     is_validator_subclass,
#     create,
#     checker,
#     check_function,
#     check_async_function
# )
# from .src.utils.path import PathUtil, PathTypes
# from .src.utils.strings import StringsUtil, Encoding, StringOrBytesOrByteArray
#
# from .src.callback.base import AbstractCallbackManager
# from .src.callback.manager import CallbackManager, AsyncCallbackManager
#
# from .src.conf.base import AbstractConfReader, IReaderFactory
# from .src.conf.reader import IniReader, JsonReader
# from .src.conf.factory import ReaderFactory
#
# from .src.policy.base import AbstractExecutor, CoroOrFunction, ArgsTypes, KwargsTypes, ReturnType
# from .src.policy.executor import Executor
# from .src.policy.task import TaskFactory

# from .src.ammeter import AbstractMessageCounter, AsyncMessageCounter, MessageCounter
from . import src


# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractTimeConvert, IntOrFloat
from .unit import UnitOfTime, MulOrTruediv, unt_map

TRUEDIV = MulOrTruediv.TRUEDIV


class TimeUnt(AbstractTimeConvert):
    _unt = unt_map(TRUEDIV)

    def __init__(self, unit_of_time: UnitOfTime):
        super().__init__(unit_of_time)

    def _convert(self, time: IntOrFloat) -> float:
        return self._get_unt().convert(float(time))

    def _get_unt(self):
        return self._unt.get(self.unit_of_time, UnitOfTime.MILLISECOND)



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .factory import ValidatorFactory


def check_function(value):
    validator = ValidatorFactory.create_func_validator(
        f"value need function or Callable, but got {type(value).__name__}"
    )
    return validator.is_validate(value)


def check_async_function(value):
    validator = ValidatorFactory.create_afunc_validator(
        f"value need async function, but got {type(value).__name__}"
    )
    return validator.is_validate(value)



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .manager import ListCallbackManager, SyncCallbackManager, AsyncCallbackManager


def func_manager():
    return SyncCallbackManager(ListCallbackManager())


def afunc_manager():
    return AsyncCallbackManager(ListCallbackManager())



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from asyncio import AbstractEventLoop
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Coroutine, Callable

from ..is_ import is_validate
from .base import ITaskExecutor, CoroOrFunction, ReturnType
from .fut import FutConvert


class CoroTaskExecutor(ITaskExecutor):

    def __init__(self, coro_or_func: Coroutine):
        self._coro = coro_or_func

    async def execute(self, *args, **kwargs) -> ReturnType:
        return await self._coro

    def __str__(self):
        return f"{self.__class__.__name__}(coro={self._coro})"


class CoroFunctionTaskExecutor(ITaskExecutor):

    def __init__(self, coro_or_func: Callable):
        self._afunc = coro_or_func

    async def execute(self, *args, **kwargs) -> ReturnType:
        return await self._afunc(*args, **kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}(coro={self._afunc})"


class FunctionTaskExecutor(ITaskExecutor):

    def __init__(self,
                 coro_or_func: Callable, *,
                 loop: AbstractEventLoop = None,
                 max_workers: int = None):
        self._func = coro_or_func
        self._convert = FutConvert(
            loop, ThreadPoolExecutor(max_workers=max_workers)
        )

    async def execute(self, *args, **kwargs) -> ReturnType:
        # sync function to coroutine
        sync = self._get_coro_or_func(**kwargs)
        fut = self._to_future(sync, *args)
        return await fut

    def _to_future(self, sync, *args):
        return self._convert.convert(sync, *args)

    def _get_coro_or_func(self, **kwargs):
        """
        保留关键字参数
        :return:
        """
        return partial(self._func, **kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}(func={self._func}, convert={self._convert})"


class TaskExecutor(ITaskExecutor):

    def __init__(self,
                 coro_or_func: CoroOrFunction,
                 *,
                 eloop: AbstractEventLoop = None,
                 max_workers: int = None):

        if is_validate.is_coroutine(coro_or_func):
            self._executor = CoroTaskExecutor(coro_or_func)

        elif is_validate.is_async_function(coro_or_func):
            self._executor = CoroFunctionTaskExecutor(coro_or_func)

        elif is_validate.is_function(coro_or_func):
            self._executor = FunctionTaskExecutor(
                coro_or_func,
                eloop=eloop, max_workers=max_workers
            )
        else:
            raise ValueError(
                f"executed need Coroutine, Task, async function, or function, but got {type(coro_or_func).__name__}"
            )

    @property
    def executor(self):
        return self._executor

    async def execute(self, *args, **kwargs) -> ReturnType:
        return await self.executor.execute(*args, **kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}(executor={self.executor})"



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from asyncio import Task
from typing import Callable, Coroutine, Dict, TypeVar, Tuple, Optional, Union

ArgsTypes = Optional[Tuple]
KwargsTypes = Optional[Dict]
CoroOrFunction = Union[Callable, Coroutine]
ReturnType = TypeVar("ReturnType")


class ITaskExecutor(metaclass=ABCMeta):

    @abstractmethod
    async def execute(self, *args, **kwargs) -> ReturnType:
        pass


class ITaskFactory(metaclass=ABCMeta):

    @abstractmethod
    def create(self) -> Task:
        pass



# @Time     : 2021/6/2
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from asyncio import Task, create_task, get_event_loop
from typing import Callable, Coroutine, Dict, TypeVar, Tuple, Union

from ..is_ import is_validate

ArgsTypes = Union[Tuple, None]
KwargsTypes = Union[Dict, None]
CoroOrFunction = Union[Callable, Coroutine]
ReturnType = TypeVar("ReturnType")


class ICoroutineAdapters(object):
    """
    coroutine adapters
    """

    def to_coroutine(self, *args, **kwargs) -> Coroutine:
        raise NotImplementedError("NotImplemented .to_coroutine(*args, **kwargs) -> Coroutine")


class AbstractExecutor(metaclass=ABCMeta):

    def __init__(self,
                 coro_or_func: CoroOrFunction,
                 args: ArgsTypes = None,
                 kwargs: KwargsTypes = None):
        self._coro_or_func = coro_or_func
        self._args = () if not args else args
        self._kwargs = {} if not kwargs else kwargs

    @property
    def coro_or_func(self) -> CoroOrFunction:
        return self._coro_or_func

    @abstractmethod
    async def execute(self) -> ReturnType:
        raise NotImplementedError("NotImplemented .execute(*args, **kwargs) -> ReturnType")


class AbstractTaskFactory(metaclass=ABCMeta):

    def __init__(self, coro_or_func: CoroOrFunction, args: ArgsTypes = None, kwargs: KwargsTypes = None):
        self._coro_or_func = coro_or_func
        self._args = () if not args else args
        self._kwargs = {} if not kwargs else kwargs

    @property
    def coro_or_func(self) -> CoroOrFunction:
        return self._coro_or_func

    @abstractmethod
    def create_task(self) -> Task:
        raise NotImplementedError("NotImplemented .create_task() -> Task")

    def _create_by_coro(self, coro):
        if not is_validate.is_coroutine(coro):
            raise ValueError(
                f"coro need Coroutine, but got {type(coro).__name__}"
            )
        return create_task(coro)


class IAsyncTask(object):
    """
    task interface
    """

    @property
    def coro_or_func(self) -> CoroOrFunction:
        raise NotImplementedError("NotImplemented .coro_or_func -> CoroOrFunction")

    def create_task(self, *args, **kwargs) -> Task:
        raise NotImplementedError("NotImplemented .create_task(*args, **kwargs) -> Task")


def _get_event_loop():
    return get_event_loop()



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from functools import partial

from .base import AbstractValidator
from .func import FunctionValidator, AsyncFunctionValidator


class AbstractValidatorFactory(metaclass=ABCMeta):

    @classmethod
    def create_validator(cls, subclass: AbstractValidator, message: str, **kwargs) -> AbstractValidator:
        factory = cls._factory(subclass, message)
        return factory(**kwargs)

    @classmethod
    def _is_subclass(cls, klass):
        return issubclass(klass, AbstractValidator)

    @classmethod
    def _factory(cls, subclass: AbstractValidator, message: str):
        if not cls._is_subclass(subclass):
            raise ValueError(f"need a {AbstractValidator.__name__} subclass, but got {type(subclass).__name__}")
        return partial(subclass, message)

    @classmethod
    @abstractmethod
    def create_func_validator(cls, message: str) -> AbstractValidator:
        pass

    @classmethod
    @abstractmethod
    def create_afunc_validator(cls, message: str) -> AbstractValidator:
        pass


class ValidatorFactory(AbstractValidatorFactory):

    @classmethod
    def create_func_validator(cls, message: str) -> FunctionValidator:
        return cls.create_validator(
            FunctionValidator, message
        )

    @classmethod
    def create_afunc_validator(cls, message: str) -> AsyncFunctionValidator:
        return cls.create_validator(
            AsyncFunctionValidator, message
        )



# @Time     : 2021/6/2
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import AbstractEventLoop
from .base import AbstractExecutor, CoroOrFunction, ArgsTypes, KwargsTypes, ReturnType, _get_event_loop
from .adapters import CoroutineAdapters
from ..validator.is_validators import is_coroutine, is_async_function, is_function


class CoroExecutor(AbstractExecutor):

    async def execute(self) -> ReturnType:
        return await self.coro_or_func


class CoroFunctionExecutor(AbstractExecutor):

    async def execute(self) -> ReturnType:
        args, kwargs = self._args, self._kwargs
        return await self.coro_or_func(*args, **kwargs)


class FunctionExecutor(AbstractExecutor):

    def __init__(self,
                 loop: AbstractEventLoop,
                 *,
                 coro_or_func: CoroOrFunction,
                 args: ArgsTypes = None,
                 kwargs: KwargsTypes = None,
                 max_workers: int = None):
        super().__init__(coro_or_func, args, kwargs)
        self._eloop = loop
        self._max_workers = max_workers

    @property
    def eloop(self):
        return self._eloop

    async def execute(self) -> ReturnType:
        # sync function to coroutine
        fut = self._to_future(
            self.eloop,
            func=self.coro_or_func,
            args=self._args,
            kwargs=self._kwargs,
            max_workers=self._max_workers
        )
        return await fut

    def _to_future(self,
                   loop: AbstractEventLoop,
                   *,
                   func,
                   args: tuple,
                   kwargs: dict,
                   max_workers: int = None):
        adapter = CoroutineAdapters(loop, func, max_workers=max_workers)
        return adapter.to_coroutine(*args, **kwargs)


class Executor(AbstractExecutor):

    def __init__(self,
                 coro_or_func: CoroOrFunction,
                 args: ArgsTypes = None,
                 kwargs: KwargsTypes = None,
                 max_workers: int = None):
        super().__init__(coro_or_func, args, kwargs)
        self._coro_or_func = coro_or_func
        self._max_workers = max_workers

    @property
    def coro_or_func(self) -> CoroOrFunction:
        return self._coro_or_func

    @property
    def max_workers(self):
        return self._max_workers

    async def execute(self) -> ReturnType:
        coro_or_func = self.coro_or_func
        if is_coroutine(coro_or_func):
            return await self._use_coro_exec(coro_or_func)
        elif is_async_function(coro_or_func):
            return await self._use_coro_func_exec(coro_or_func)
        elif is_function(coro_or_func):
            return await self._use_func_exec(coro_or_func)
        else:
            raise ValueError(
                f"executed need Coroutine, Task, async function, or function, but got {type(coro_or_func).__name__}"
            )

    async def _use_coro_exec(self, coro_or_func: CoroOrFunction):
        executor = CoroExecutor(
            coro_or_func=coro_or_func,
            args=self._args,
            kwargs=self._kwargs
        )
        return await executor.execute()

    async def _use_coro_func_exec(self, coro_or_func: CoroOrFunction):
        executor = CoroFunctionExecutor(
            coro_or_func=coro_or_func,
            args=self._args,
            kwargs=self._kwargs
        )
        return await executor.execute()

    async def _use_func_exec(self, coro_or_func: CoroOrFunction):
        event_loop = _get_event_loop()
        executor = FunctionExecutor(
            event_loop,
            coro_or_func=coro_or_func,
            args=self._args,
            kwargs=self._kwargs,
            max_workers=self.max_workers
        )
        return await executor.execute()



# @Time     : 2021/5/31
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from typing import Optional

from ..utils import PathUtil, PathTypes, EncodingType
from .base import AbstractConfReader
from .reader import IniReader, JsonReader


class IReaderFactory(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def create(cls, file: PathTypes, encoding: Optional[EncodingType] = None) -> AbstractConfReader:
        raise NotImplementedError("NotImplemented .create() -> AbstractConfReader")


class ReaderFactory(IReaderFactory):

    @classmethod
    def create(cls, file: PathTypes, encoding: Optional[EncodingType] = None):
        if cls.is_ini(file):
            return IniReader(file, encoding)
        elif cls.is_json(file):
            return JsonReader(file, encoding)
        else:
            raise ValueError(f"file ext name error, {file}")

    @staticmethod
    def is_json(file):
        return PathUtil.is_endswith_ext(file, [".json"])

    @staticmethod
    def is_ini(file):
        return PathUtil.is_endswith_ext(file, [".ini"])



# @Time     : 2021/8/13
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from ..time_ import UnitOfTime, TimeType
from .base import Function
from .meter import AsyncMessageMeter


def meter(timeout: TimeType, unt: UnitOfTime = UnitOfTime.MILLISECOND, *, callback: Function = None):
    return AsyncMessageMeter(timeout, unt, callback=callback)



# @Time     : 2021/6/1
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from abc import ABCMeta, abstractmethod
from operator import not_
from typing import Callable, Iterable, List, Union, TypeVar

from ..is_ import is_validate
from .base import AbstractValidator

ValidatorTypes = Union[AbstractValidator, Callable]
ValidatorList = List[ValidatorTypes]
Validated = TypeVar("Validated")


class AbstractValidatorManager(metaclass=ABCMeta):

    @property
    @abstractmethod
    def validators(self) -> Iterable[ValidatorTypes]:
        pass

    @property
    @abstractmethod
    def length(self) -> int:
        pass

    @abstractmethod
    def add(self, validator: ValidatorTypes) -> int:
        pass

    @abstractmethod
    def remove(self, validator: ValidatorTypes) -> int:
        pass

    def empty(self):
        return not_(self.length)

    def is_validate(self, value: Validated) -> bool:
        results = []
        append = results.append
        for validator in self:
            append(results)
        return all(results)

    def __iter__(self):
        if not self.empty():
            for _, item in enumerate(self.validators):
                yield item


class ListValidatorManager(AbstractValidatorManager):

    def __init__(self):
        self._validators: ValidatorList = []

    @property
    def validators(self):
        return self._validators

    @property
    def length(self) -> int:
        return len(self.validators)

    def add(self, validator):
        if not self._is_validator(validator):
            return self.length

        if not self._is_exists(validator):
            self._validators.append(validator)

        return self.length

    def remove(self, validator):
        if self.empty():
            return self.length
        idx = self._find(validator)
        if idx > -1:
            self._validators.pop(idx)
        return self.length

    def _find(self, validator: ValidatorTypes) -> int:
        if not self._is_exists(validator):
            return -1
        return self._validators.index(validator)

    def _is_validator(self, value) -> bool:
        return any([
            is_validate.is_function(value),
            isinstance(value, AbstractValidator)
        ])

    def _is_exists(self, validator: ValidatorTypes) -> bool:
        if self.empty():
            return False
        return validator in self.validators



# @Time     : 2021/6/2
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import AbstractEventLoop, get_event_loop
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Callable

from .base import ICoroutineAdapters


class CoroutineAdapters(ICoroutineAdapters):
    """
    function to coroutine
    """

    def __init__(self,
                 loop: AbstractEventLoop,
                 func: Callable,
                 *, max_workers: int = None):
        self._eloop = loop
        self._func = func
        self._executor = ThreadPoolExecutor(max_workers=max_workers)

    @property
    def func(self):
        return self._func

    @property
    def executor(self):
        return self._executor

    def to_coroutine(self, *args, **kwargs):
        function = self._get_function(**kwargs)
        return self._create_future(function, *args)

    def _create_future(self, sync, *args):
        """
        将 sync func -> fut
        :param sync:
        :param args:
        :return:
        """
        event_loop = self._get_event_loop()
        fut = event_loop.run_in_executor(self.executor, sync, *args)
        return fut

    def _get_function(self, **kwargs):
        """
        保留关键字参数
        :param kwargs:
        :return:
        """
        return partial(self.func, **kwargs)

    def _get_event_loop(self):
        """
        获取 event loop
        :return:
        """
        if self._eloop is None:
            return get_event_loop()
        return self._eloop
